# -*- encoding:utf-8 -*-
"""
    示例仓位管理：atr仓位管理模块
"""

from ultron.ump.beta.position_base import PositionBase
"""
    默认0.1即10% 外部可通过如：abupy.beta.atr.g_atr_pos_base = 0.01修改仓位基础配比
    需要注意外部其它自定义仓位管理类不要随意使用模块全局变量，AbuAtrPosition特殊因为注册
    在ABuEnvProcess中在多进程启动时拷贝了模块全局设置内存
"""
g_atr_pos_base = 0.1


class AtrPosition(PositionBase):
    """示例atr仓位管理类"""

    s_atr_base_price = 15  # best fit wide: 12-20
    s_std_atr_threshold = 0.5  # best fit wide: 0.3-0.65

    def fit_position(self, factor_object):
        """
        fit_position计算的结果是买入多少个单位（股，手，顿，合约）
        计算：（常数价格 ／ 买入价格）＊ 当天交易日atr21
        :param factor_object: ABuFactorBuyBases实例对象
        :return: 买入多少个单位（股，手，顿，合约）
        """
        std_atr = (self.atr_base_price / self.bp) * self.kl_pd_buy['atr21']
        """
            对atr 进行限制 避免由于股价波动过小，导致
            atr小，产生大量买单，实际上针对这种波动异常（过小，过大）的股票
            需要有其它的筛选过滤策略, 选股的时候取0.5，这样最大取两倍g_atr_pos_base
        """
        atr_wv = self.std_atr_threshold if std_atr < self.std_atr_threshold else std_atr
        # 计算出仓位比例
        atr_pos = self.atr_pos_base / atr_wv
        # 最大仓位限制
        atr_pos = self.pos_max if atr_pos > self.pos_max else atr_pos
        # 结果是买入多少个单位（股，手，顿，合约）
        return self.read_cash * atr_pos / self.bp * self.deposit_rate

    def _init_self(self, **kwargs):
        """atr仓位控制管理类初始化设置"""
        self.atr_base_price = kwargs.pop('atr_base_price',
                                         AtrPosition.s_atr_base_price)
        self.std_atr_threshold = kwargs.pop('std_atr_threshold',
                                            AtrPosition.s_std_atr_threshold)
        self.atr_pos_base = kwargs.pop('atr_pos_base', g_atr_pos_base)



# -*- encoding:utf-8 -*-
"""
    相关性降维
"""
import pdb
import numpy as np
from enum import Enum
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress
from ultron.ump.similar.corrcoef import corr_xy, corr_matrix, ECoreCorrType


class FCorrType(Enum):
    """
        FCorrType:因子相关性计算
    """
    """横截面相关系数计算"""
    F_CS_CORR = 'cs_corr'
    """时序相关系数计算"""
    F_TS_CORR = 'ts_corr'


class Corrcoef(object):

    def __init__(self, features=None, thresh=0.7, method='squares'):
        self._features = features
        self._thresh = thresh
        self._method = method
        self._filter = []

    def _corr_xy(self, x, y, similar_type, **kwargs):
        return corr_xy(x, y, similar_type=similar_type, **kwargs)

    def _ts_corr(self, factors_data, similar_type, **kwargs):
        grouped = factors_data.set_index(['trade_date'
                                          ]).groupby(level=['trade_date'])
        p = 0
        with Progress(len(self._features) - 1, 0, 'TS CORR') as pg:
            for i in range(0, len(self._features) - 1):
                p += 1
                for j in range(1, len(self._features)):
                    x_col = self._features[i]
                    y_col = self._features[j]
                    if x_col in self._filter or y_col in self._filter or x_col == y_col:
                        continue
                    res = []
                    for k, g in grouped:
                        corr = corr_xy(g[x_col], g[y_col], similar_type,
                                       **kwargs)
                        res.append(corr)
                    corr_mean = np.array(res).mean()
                    kd_logger.info("{0},{1} corr:{2}".format(
                        x_col, y_col, corr_mean))
                    if abs(corr_mean) > self._thresh:
                        self._filter.append(y_col)
                pg.show(p)

    def _cs_corr(self, factors_data, similar_type, **kwargs):
        p = 0
        with Progress(len(self._features) - 1, 0, 'CS CORR') as pg:
            for i in range(0, len(self._features) - 1):
                p += 1
                for j in range(1, len(self._features)):
                    x_col = self._features[i]
                    y_col = self._features[j]
                    if x_col in self._filter or y_col in self._filter or x_col == y_col:
                        continue
                    corr = self._corr_xy(factors_data[x_col],
                                         factors_data[y_col], similar_type,
                                         **kwargs)
                    kd_logger.info("{0},{1} corr:{2}".format(
                        x_col, y_col, corr))
                    if abs(corr) > self._thresh:
                        self._filter.append(y_col)
                pg.show(p)

    def run(self, factors_data, similar_type, **kwargs):
        if self._method == FCorrType.F_CS_CORR:
            self._cs_corr(factors_data=factors_data,
                          similar_type=similar_type,
                          **kwargs)
        elif self._method == FCorrType.F_TS_CORR:
            self._ts_corr(factors_data=factors_data,
                          similar_type=similar_type,
                          **kwargs)
        return factors_data.drop(self._filter, axis=1)



# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._Months cimport Months as mt
cdef public enum Months:
    January = 1
    February = 2
    March = 3
    April = 4
    May = 5
    June = 6
    July = 7
    August = 8
    September = 9
    October = 10
    November = 11
    December = 12


# -*- encoding:utf-8 -*-
"""
MACD

MACD称为指数平滑异动移动平均线，是从双指数移动平均线发展而来的，由快的加权移动均线（EMA12）减去慢的加权移动均线（EMA26）
得到DIF，再用DIF - (快线-慢线的9日加权移动均线DEA）得到MACD柱。MACD的意义和双移动平均线基本相同，即由快、慢均线的离散、
聚合表征当前的多空状态和股价可能的发展变化趋势，但阅读起来更方便。当MACD从负数转向正数，是买的信号。当MACD从正数转向负数，
是卖的信号。当MACD以大角度变化，表示快的移动平均线和慢的移动平均线的差距非常迅速的拉开，代表了一个市场大趋势的转变。

"""

import numpy as np
import pandas as pd

from ultron.ump.indicator.base import g_calc_type, ECalcType
from ultron.kdutils import scaler as scaler_util
from ultron.kdutils.decorator import catch_error
from ultron.ump.core.helper import pd_ewm_mean


# noinspection PyUnresolvedReferences
def _calc_macd_from_ta(price, fast_period=12, slow_period=26, signal_period=9):
    """
    使用talib计算macd, 即透传talib.MACD计算结果
    :param price: 收盘价格序列，pd.Series或者np.array
    :param fast_period: 快的加权移动均线线, 默认12，即EMA12
    :param slow_period: 慢的加权移动均线, 默认26，即EMA26
    :param signal_period: dif的指数移动平均线，默认9
    """

    import talib
    if isinstance(price, pd.Series):
        price = price.values

    dif, dea, bar = talib.MACD(price,
                               fastperiod=fast_period,
                               slowperiod=slow_period,
                               signalperiod=signal_period)
    return dif, dea, bar


def _calc_macd_from_pd(price, fast_period=12, slow_period=26, signal_period=9):
    """
    通过macd公式手动计算macd
    :param price: 收盘价格序列，pd.Series或者np.array
    :param fast_period: 快的加权移动均线线, 默认12，即EMA12
    :param slow_period: 慢的加权移动均线, 默认26，即EMA26
    :param signal_period: dif的指数移动平均线，默认9
    """

    if isinstance(price, pd.Series):
        price = price.values

    # 快的加权移动均线
    ewma_fast = pd_ewm_mean(price, span=fast_period)
    # 慢的加权移动均线
    ewma_slow = pd_ewm_mean(price, span=slow_period)
    # dif = 快线 - 慢线
    dif = ewma_fast - ewma_slow
    # dea = dif的9日加权移动均线
    dea = pd_ewm_mean(dif, span=signal_period)
    bar = (dif - dea)
    return dif, dea, bar


"""通过在NDBase中尝试import talib来统一确定指标计算方式"""
calc_macd = _calc_macd_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_macd_from_ta



# -*- encoding:utf-8 -*-
"""
    全局环境配置模块
"""
import os
from os import path
from enum import Enum
from urllib.request import urlretrieve
from ultron.utilities.logger import kd_logger
"""是否开启 example 环境，默认关闭False"""
_g_enable_example_env = False
"""当前用户路径"""
root_drive = path.expanduser('~')

g_project_root = path.join(root_drive, 'ultron')

g_base_path = os.path.join(g_project_root, 'base')

g_project_data = os.path.join(g_project_root, 'rom/sandbox')

default_sandbox_url = 'https://ultronsandbox.oss-cn-hangzhou.aliyuncs.com/sandbox.zip'
default_base_url = 'https://ultronsandbox.oss-cn-hangzhou.aliyuncs.com/base.zip'

g_sandbox_url = default_sandbox_url if 'ULTRON_DATA_URL' not in os.environ else os.environ[
    'ULTRON_DATA_URL']

g_base_url = default_base_url if 'ULTRON_BASE_URL' not in os.environ else os.environ[
    'ULTRON_BASE_URL']

if 'ULTRON_DATA' in os.environ:
    g_project_data = os.path.join(g_project_data, os.environ['ULTRON_DATA'])

g_plt_figsize = (14, 7)


# TODO 缩短 E_DATA_FETCH_NORMAL－>NORMAL
class EMarketDataFetchMode(Enum):
    """
        金融时间数据获取模式
    """
    """普通模式，尽量从本地获取数据，本地数据不满足的情况下进行网络请求"""
    E_DATA_FETCH_NORMAL = 0
    """强制从本地获取数据，本地数据不满足的情况下，返回None"""
    E_DATA_FETCH_FORCE_LOCAL = 1
    """强制从网络获取数据，不管本地数据是否满足"""
    E_DATA_FETCH_FORCE_NET = 2


def init_plot_set():
    """全局plot设置"""
    import seaborn as sns
    sns.set_context('notebook', rc={'figure.figsize': g_plt_figsize})
    sns.set_style("darkgrid")

    import matplotlib
    # conda 5.0后需要添加单独matplotlib的figure设置否则pandas的plot size不生效
    matplotlib.rcParams['figure.figsize'] = g_plt_figsize


def init_data():
    if not os.path.exists(g_base_path):
        data_base_zip = os.path.join(g_project_root, "base.zip")
        if not os.path.exists(g_project_root):
            os.mkdir(g_project_root)
        kd_logger.info("download base data")
        urlretrieve(g_base_url, data_base_zip)
        try:
            from zipfile import ZipFile
            zip_csv = ZipFile(data_base_zip, "r")
            unzip_dir = os.path.join(g_project_root, "./")
            for csv in zip_csv.namelist():
                zip_csv.extract(csv, unzip_dir)
            zip_csv.close()
        except Exception as e:
            # 解压测试数据zip失败，就不开启测试数据模式了
            kd_logger.error('data env failed! e={}'.format(e))
            return


def enable_example_env(show_log=True):
    if not os.path.exists(g_project_data):
        data_example_zip = os.path.join(g_project_root, "sandbox.zip")
        if not os.path.exists(g_project_root):
            os.mkdir(g_project_root)
        kd_logger.info("download sandbox data")
        urlretrieve(g_sandbox_url, data_example_zip)
        try:
            from zipfile import ZipFile
            zip_csv = ZipFile(data_example_zip, "r")
            unzip_dir = os.path.join(g_project_root, "rom/")
            for csv in zip_csv.namelist():
                zip_csv.extract(csv, unzip_dir)
            zip_csv.close()
        except Exception as e:
            # 解压测试数据zip失败，就不开启测试数据模式了
            kd_logger.error('example env failed! e={}'.format(e))
            return

    global _g_enable_example_env, g_data_fetch_mode
    _g_enable_example_env = True

    g_data_fetch_mode = EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL

    init_plot_set()
    if show_log:
        kd_logger.info(
            'enable example env will only read {0}'.format(g_project_data))



# -*- encoding:utf-8 -*-
from collections import namedtuple
from ultron.kdutils import date as date_utils
from ultron.ump.core import env
from ultron.utilities.logger import kd_logger
from ultron.ump.market.benchmark import Benchmark
from ultron.ump.trade.kl_manager import KLManager
from ultron.ump.trade.capital import Capital
from ultron.ump.alpha.pick_time_master import PickTimeMaster


# noinspection PyClassHasNoInit
class ResultTuple(
        namedtuple('AbuResultTuple',
                   ('orders_pd', 'action_pd', 'capital', 'benchmark'))):
    """
        使用abu.run_loop_back返回的nametuple对象：

        orders_pd：回测结果生成的交易订单构成的pd.DataFrame对象
        action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        capital:   资金类AbuCapital实例化对象
        benchmark: 交易基准对象，AbuBenchmark实例对象
    """
    __slots__ = ()

    def __repr__(self):
        """打印对象显示：orders_pd.info, action_pd.info, capital, benchmark"""
        return "orders_pd:{}\naction_pd:{}\ncapital:{}\nbenchmark:{}".format(
            self.orders_pd.info(), self.action_pd.info(), self.capital,
            self.benchmark)


def run_loop_back(read_cash,
                  buy_factors,
                  sell_factors,
                  pick_kl_pd_dict,
                  choice_symbols=None,
                  commission_dict=None,
                  n_process_kl=None,
                  n_process_pick=None,
                  benchmark_kl_pd=None):
    """
    封装执行择时，选股回测。

    推荐在使用abu.run_loop_back()函数进行全市场回测前使用abu.run_kl_update()函数首先将数据进行更新，
    在run_kl_update()中它会首选强制使用网络数据进行更新，在更新完毕后，更改数据获取方式为本地缓存，
    使用abu.run_kl_update()的好处是将数据更新与策略回测分离，在运行效率及问题排查上都会带来正面的提升

    :param read_cash: 初始化资金额度，eg：1000000
    :param buy_factors: 回测使用的买入因子策略序列，
                    eg：
                        buy_factors = [{'xd': 60, 'class': AbuFactorBuyBreak},
                                       {'xd': 42, 'class': AbuFactorBuyBreak}]
    :param sell_factors: 回测使用的卖出因子序列，
                    eg:
                        sell_factors = [{'stop_loss_n': 0.5, 'stop_win_n': 3.0, 'class': AbuFactorAtrNStop},
                                        {'pre_atr_n': 1.0, 'class': AbuFactorPreAtrNStop},
                                        {'close_atr_n': 1.5, 'class': AbuFactorCloseAtrNStop},]
    :param stock_picks: 回测使用的选股因子序列：
                    eg:
                        stock_pickers = [{'class': AbuPickRegressAngMinMax,
                                          'threshold_ang_min': 0.0, 'reversed': False},
                                         {'class': AbuPickStockPriceMinMax,
                                          'threshold_price_min': 50.0,
                                          'reversed': False}]
    :param choice_symbols: 备选股票池, 默认为None，即使用abupy.env.g_market_target的市场类型进行全市场回测，
                           为None的情况下为symbol序列
                    eg:
                        choice_symbols = ['usNOAH', 'usSFUN', 'usBIDU', 'usAAPL', 'usGOOG',
                                          'usTSLA', 'usWUBA', 'usVIPS']
    :param n_folds: int, 回测n_folds年的历史数据
    :param start: 回测开始的时间, str对象, eg: '2013-07-10'
    :param end: 回测结束的时间, str对象 eg: '2016-07-26'
    :param commission_dict: 透传给Capital，自定义交易手续费的时候时候。
                    eg：
                        def free_commission(trade_cnt, price):
                            # 免手续费
                            return 0
                        commission_dict = {'buy_commission_func': free_commission,
                                         'sell_commission_func': free_commission}
                        Capital(read_cash, benchmark, user_commission_dict=commission_dict)

    :param n_process_kl: 金融时间序列数据收集启动并行的进程数，默认None, 内部根据cpu数量分配
    :param n_process_pick: 择时与选股操作启动并行的进程数，默认None, 内部根据cpu数量分配
    :return: (ResultTuple对象, KLManager对象)
    """
    #if start is not None and end is not None and date_utils.date_str_to_int(
    #        end) - date_utils.date_str_to_int(start) <= 0:
    #    kd_logger.info('end date <= start date!!')
    #    return None, None

    benchmark = Benchmark(benchmark_kl_pd=benchmark_kl_pd)

    capital = Capital(read_cash,
                      benchmark,
                      user_commission_dict=commission_dict)

    win_to_one = choice_symbols is not None and len(
        choice_symbols) < 20 and not env.g_is_mac_os and env.g_cpu_cnt <= 4

    if n_process_pick is None:
        # 择时，选股并行操作的进程等于cpu数量, win_to_one满足情况下1个
        n_process_pick = 1 if win_to_one else env.g_cpu_cnt
    if n_process_kl is None:
        # mac系统下金融时间序列数据收集启动两倍进程数, windows只是进程数量，win_to_one满足情况下1个
        n_process_kl = 1 if win_to_one else env.g_cpu_cnt * 2 if env.g_is_mac_os else env.g_cpu_cnt

    kl_pd_manager = KLManager(benchmark, capital)
    kl_pd_manager.set_pick_time(pick_kl_pd_dict)

    orders_pd, action_pd, all_fit_symbols_cnt = PickTimeMaster.do_symbols_with_same_factors_process(
        choice_symbols,
        benchmark,
        buy_factors,
        sell_factors,
        capital,
        kl_pd_manager=kl_pd_manager,
        n_process_kl=n_process_kl,
        n_process_pick_time=n_process_pick)

    result_tuple = ResultTuple(orders_pd, action_pd, capital, benchmark)

    return result_tuple, kl_pd_manager



# -*- coding: utf-8 -*-

import numpy as np
import numba as nb
from typing import Tuple
from typing import Union
from typing import Dict
from . utilities import groupby


def neutralize(x: np.ndarray,
               y: np.ndarray,
               groups: np.ndarray=None,
               detail: bool=False,
               weights: np.ndarray = None) \
        -> Union[np.ndarray, Tuple[np.ndarray, Dict]]:

    if y.ndim == 1:
        y = y.reshape((-1, 1))

    if weights is None:
        weights = np.ones(len(y), dtype=float)

    output_dict = {}

    if detail:
        exposure = np.zeros(x.shape + (y.shape[1],))
        explained = np.zeros(x.shape + (y.shape[1],))
        output_dict['exposure'] = exposure
        output_dict['explained'] = explained

    if groups is not None:
        res = np.zeros(y.shape)
        index_diff, order = groupby(groups)
        start = 0
        if detail:
            for diff_loc in index_diff:
                curr_idx = order[start:diff_loc + 1]
                curr_x, b = _sub_step(x, y, weights, curr_idx, res)
                exposure[curr_idx, :, :] = b
                explained[curr_idx] = ls_explain(curr_x, b)
                start = diff_loc + 1
        else:
            for diff_loc in index_diff:
                curr_idx = order[start:diff_loc + 1]
                _sub_step(x, y, weights, curr_idx, res)
                start = diff_loc + 1
    else:
        try:
            b = ls_fit(x, y, weights)
        except np.linalg.linalg.LinAlgError:
            b = ls_fit_pinv(x, y, weights)

        res = ls_res(x, y, b)

        if detail:
            explained[:, :, :] = ls_explain(x, b)
            exposure[:] = b

    if output_dict:
        return res, output_dict
    else:
        return res


def _sub_step(x, y, w, curr_idx, res) -> Tuple[np.ndarray, np.ndarray]:
    curr_x, curr_y, curr_w = x[curr_idx], y[curr_idx], w[curr_idx]
    try:
        b = ls_fit(curr_x, curr_y, curr_w)
    except np.linalg.linalg.LinAlgError:
        b = ls_fit_pinv(curr_x, curr_y, curr_w)
    res[curr_idx] = ls_res(curr_x, curr_y, b)
    return curr_x, b


@nb.njit(nogil=True, cache=True)
def ls_fit(x: np.ndarray, y: np.ndarray, w: np.ndarray) -> np.ndarray:
    x_bar = x.T * w
    b = np.linalg.solve(x_bar @ x, x_bar @ y)
    return b


@nb.njit(nogil=True, cache=True)
def ls_fit_pinv(x: np.ndarray, y: np.ndarray, w: np.ndarray) -> np.ndarray:
    x_bar = x.T * w
    b = np.linalg.pinv(x_bar @ x) @ x_bar @ y
    return b


@nb.njit(nogil=True, cache=True)
def ls_res(x: np.ndarray, y: np.ndarray, b: np.ndarray) -> np.ndarray:
    return y - x @ b


@nb.njit(nogil=True, cache=True)
def ls_explain(x: np.ndarray, b: np.ndarray) -> np.ndarray:
    m, n = b.shape
    return b.reshape((1, m, n)) * x.reshape((-1, m, 1))


# -*- encoding:utf-8 -*-
"""示例ump边裁特征走势拟合角度模块"""

from ultron.ump.trade.ml_feature import FeatureDeg, FeatureDegExtend
from ultron.ump.ump.base import ump_edge_make_xy, BuyUmpMixin
from ultron.ump.ump.edge_base import EdgeBase
from ultron.ump.model.principles import Principles


class DegFiter(Principles):
    """
        只需要实现make_xy，且使用ump_edge_make_xy装饰
        在边裁__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_edge_make_xy
    def make_xy(self, **kwarg):
        """
         make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
         order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
         make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

        :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
        :return: self.order_has_ret中使用filter选取特征列形成deg_df
                 ump_edge_make_xy装饰器在make_xy返回deg_df后做转换matrix，形成x，y等工作
        """

        filter_list = ['profit', 'profit_cg']
        cols = FeatureDeg().get_feature_ump_keys(ump_cls=EdgeDeg)
        # filter_list: ['profit', 'profit_cg', 'buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        filter_list.extend(cols)

        # noinspection PyUnresolvedReferences
        deg_df = self.order_has_ret.filter(filter_list)
        """
            eg: deg_df
                           profit  profit_cg  buy_deg_ang42  buy_deg_ang252  \
            2014-09-24  -22618.04    -0.0566          3.378           3.458
            2014-10-24  -29690.28    -0.0742          0.191           2.889
            2014-10-29   18959.19     0.0542         -2.026          16.689
            2014-10-29  148209.36     0.5022         -3.427         -11.956
            2014-10-29   24867.60     0.0952         -2.915          39.469
            2014-10-29   18959.19     0.0542         -2.026          16.689
            2014-11-03    1250.80     0.0045          0.103          39.202
            2014-11-11   59888.21     0.1857          8.341          -9.450
            2014-11-12   -3578.78    -0.0140          3.963           6.595
            2014-11-26  -29085.19    -0.0946         14.052           6.061
            ...               ...        ...            ...             ...
            2016-03-14   16220.57     0.0559          4.002         -10.559
            2016-03-14  -25328.12    -0.1218          0.129          -6.649
            2016-03-30  -29858.44    -0.0863         13.121          -8.461
            2016-04-04    5373.76     0.0244          4.409         -33.097
            2016-04-13  -28044.40    -0.1159          6.603         -31.459
            2016-04-14  -18645.93    -0.0467          4.611          18.428
            2016-04-15  -32484.79    -0.1149          4.238         -13.247
            2016-04-15  -32484.79    -0.1149          4.238         -13.247
            2016-04-29     290.96     0.0007          1.445          16.266
            2016-04-29     290.96     0.0007          1.445          16.266

                        buy_deg_ang60  buy_deg_ang21
            2014-09-24          3.458          1.818
            2014-10-24          2.809         -1.089
            2014-10-29         -0.761          1.980
            2014-10-29         -8.296          6.507
            2014-10-29         -6.043          7.046
            2014-10-29         -0.761          1.980
            2014-11-03         -4.614         10.125
            2014-11-11          0.730         12.397
            2014-11-12         -7.524          6.671
            2014-11-26          7.566         12.494
            ...                   ...            ...
            2016-03-14         -7.992          9.324
            2016-03-14        -10.880          5.201
            2016-03-30          4.498          4.070
            2016-04-04         -6.281          5.618
            2016-04-13          0.191          4.457
            2016-04-14          3.134          0.733
            2016-04-15          4.693          1.162
            2016-04-15          4.693          1.162
            2016-04-29          4.615         -1.115
            2016-04-29          4.615         -1.115
        """
        return deg_df


class EdgeDeg(EdgeBase, BuyUmpMixin):
    """边裁走势拟合角度特征类，UltronUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        边裁走势拟合角度特征keys：['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        """
        return FeatureDeg().get_feature_keys(ump_cls=EdgeDeg)

    def get_fiter_class(self):
        """
        主裁特征走势拟合角度返回的UltronMLPd子类：UltronUmpMainDeg.UmpDegFiter
        :return: UltronUmpMainDeg.UmpDegFiter
        """
        return DegFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'deg_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'deg_edge'


class ExtendEdgeFiter(Principles):
    """角度边裁扩展类make_xy"""

    @ump_edge_make_xy
    def make_xy(self, **kwarg):
        filter_list = ['profit', 'profit_cg']
        col = FeatureDegExtend().get_feature_ump_keys(ump_cls=EegeDegExtend)
        filter_list.extend(col)
        mul_df = self.order_has_ret.filter(filter_list)
        return mul_df


class EegeDegExtend(EdgeBase, BuyUmpMixin):
    """边裁使用新的视角来决策交易，UltronUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        # 这里使用get_feature_ump_keys，只需要传递当前类名称即可，其根据是买入ump还是卖出ump返回对应特征列
        col = FeatureDegExtend().get_feature_ump_keys(ump_cls=EegeDegExtend)
        return col

    def get_fiter_class(self):
        return ExtendEdgeFiter

    @classmethod
    def class_unique_id(cls):
        return 'extend_edge_deg'



# -*- encoding:utf-8 -*-
"""示例ump边裁价格波动特征模块"""

from ultron.ump.trade.ml_feature import FeatureWave
from ultron.ump.ump.base import ump_edge_make_xy, BuyUmpMixin
from ultron.ump.ump.edge_base import EdgeBase
from ultron.ump.model.principles import Principles


class WaveFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

        在边裁__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_edge_make_xy
    def make_xy(self, **kwarg):
        """
         make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
         order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
         make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

        :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
        :return: self.order_has_ret中使用filter选取特征列形成wave_df
                 ump_edge_make_xy装饰器在make_xy返回wave_df后做转换matrix，形成x，y等工作
        """

        filter_list = ['profit', 'profit_cg']
        cols = FeatureWave().get_feature_ump_keys(ump_cls=EdgeWave)
        # filter_list=['profit', 'profit_cg', 'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        filter_list.extend(cols)
        # noinspection PyUnresolvedReferences
        wave_df = self.order_has_ret.filter(filter_list)
        """
            eg: wave_df
                           profit  profit_cg  buy_wave_score1  buy_wave_score2  \
            2014-09-24  -22618.04    -0.0566            0.287            0.234
            2014-10-24  -29690.28    -0.0742            0.596            0.488
            2014-10-29   18959.19     0.0542            0.444            0.338
            2014-10-29  148209.36     0.5022           -0.173           -0.202
            2014-10-29   24867.60     0.0952            0.031           -0.128
            2014-10-29   18959.19     0.0542            0.444            0.338
            2014-11-03    1250.80     0.0045            0.018           -0.128
            2014-11-11   59888.21     0.1857           -0.144           -0.060
            2014-11-12   -3578.78    -0.0140           -0.453           -0.505
            2014-11-26  -29085.19    -0.0946           -0.005           -0.007
            ...               ...        ...              ...              ...
            2016-03-14   16220.57     0.0559            0.928            0.941
            2016-03-14  -25328.12    -0.1218            1.209            0.891
            2016-03-30  -29858.44    -0.0863            0.470            0.630
            2016-04-04    5373.76     0.0244            0.363            0.608
            2016-04-13  -28044.40    -0.1159            0.271            0.509
            2016-04-14  -18645.93    -0.0467           -0.030            0.081
            2016-04-15  -32484.79    -0.1149            0.596            0.753
            2016-04-15  -32484.79    -0.1149            0.596            0.753
            2016-04-29     290.96     0.0007            0.743            0.840
            2016-04-29     290.96     0.0007            0.743            0.840

                        buy_wave_score3
            2014-09-24            0.218
            2014-10-24            0.449
            2014-10-29            0.329
            2014-10-29           -0.203
            2014-10-29           -0.173
            2014-10-29            0.329
            2014-11-03           -0.172
            2014-11-11            0.001
            2014-11-12           -0.509
            2014-11-26            0.015
            ...                     ...
            2016-03-14            0.948
            2016-03-14            0.788
            2016-03-30            0.702
            2016-04-04            0.743
            2016-04-13            0.651
            2016-04-14            0.170
            2016-04-15            0.800
            2016-04-15            0.800
            2016-04-29            0.918
            2016-04-29            0.918

        """
        return wave_df


class EdgeWave(EdgeBase, BuyUmpMixin):
    """边裁价格波动特征类，UltronUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        边裁价格波动特征keys：['buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        :return: ['buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        """
        return FeatureWave().get_feature_ump_keys(ump_cls=EdgeWave)

    def get_fiter_class(self):
        """
        边裁价格波动特征返回的UltronMLPd子类：UltronUmpEdgeWave.UmpWaveFiter
        :return: UltronUmpEdgeWave.UmpWaveFiter
        """
        return WaveFiter

    @classmethod
    def class_unique_id(cls):
        """
         具体ump类关键字唯一名称，类方法：return 'wave_edge'
         主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
         具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'wave_edge'



# -*- encoding:utf-8 -*-
"""示例ump价格特征模块"""

from ultron.ump.trade.ml_feature import FeaturePrice
from ultron.ump.ump.base import ump_main_make_xy, BuyUmpMixin
from ultron.ump.ump.main_base import MainBase
from ultron.ump.model.principles import Principles


class PriceFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

        在__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        """
        make_xy通过装饰器ump_main_make_xy进行二次包装
        这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
        :param kwarg: ump_main_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                      kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

        :return: self.order_has_ret中使用filter选取特征列形成price_df
                 ump_main_make_xy装饰器在make_xy返回price_df后做转换matrix，形成x，y等工作
        """

        # regex='result|buy_price_rank120|buy_price_rank90|buy_price_rank60|buy_price_rank252'
        regex = 'result|{}'.format('|'.join(
            FeaturePrice().get_feature_ump_keys(ump_cls=MainPrice)))
        # noinspection PyUnresolvedReferences
        price_df = self.order_has_ret.filter(regex=regex)
        """
            eg: price_df
                        result  buy_price_rank120  buy_price_rank90  buy_price_rank60  \
            2014-09-24       0              1.000             1.000             1.000
            2014-10-24       0              1.000             1.000             1.000
            2014-10-29       1              1.000             1.000             1.000
            2014-10-29       1              0.925             0.900             0.883
            2014-10-29       1              0.962             0.950             0.925
            2014-10-29       1              1.000             1.000             1.000
            2014-11-03       1              1.000             1.000             1.000
            2014-11-11       1              0.954             0.939             0.992
            2014-11-12       0              0.475             0.522             0.783
            2014-11-26       0              0.642             0.733             1.000
            ...            ...                ...               ...               ...
            2016-03-14       1              0.617             0.500             0.750
            2016-03-14       0              0.683             0.589             0.850
            2016-03-30       0              0.658             0.667             1.000
            2016-04-04       1              0.400             0.511             0.767
            2016-04-13       0              0.567             0.722             1.000
            2016-04-14       0              0.875             0.878             0.967
            2016-04-15       0              0.775             0.733             1.000
            2016-04-15       0              0.775             0.733             1.000
            2016-04-29       1              1.000             1.000             1.000
            2016-04-29       1              1.000             1.000             1.000

                        buy_price_rank252
            2014-09-24              1.000
            2014-10-24              1.000
            2014-10-29              1.000
            2014-10-29              0.750
            2014-10-29              0.982
            2014-10-29              1.000
            2014-11-03              1.000
            2014-11-11              0.808
            2014-11-12              0.560
            2014-11-26              0.762
            ...                       ...
            2016-03-14              0.444
            2016-03-14              0.623
            2016-03-30              0.536
            2016-04-04              0.190
            2016-04-13              0.270
            2016-04-14              0.940
            2016-04-15              0.631
            2016-04-15              0.631
            2016-04-29              1.000
            2016-04-29              1.000
        """
        return price_df


class MainPrice(MainBase, BuyUmpMixin):
    """主裁价格特征类，UltronUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        主裁价格特征keys：['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        :return: ['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        """
        return FeaturePrice().get_feature_ump_keys(ump_cls=MainPrice)

    def get_fiter_class(self):
        """
        主裁价格特征返回的UltronMLPd子类：UltronUmpMainPrice.UmpPriceFiter
        :return: UltronUmpMainPrice.UmpPriceFiter
        """
        return PriceFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'price_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'price_main'



# -*- coding: utf-8 -*-
from ultron.tradingday.Calendar import Calendar
from ultron.tradingday.Date import Date
from ultron.tradingday.Date import check_date
from ultron.tradingday.Period import Period
from ultron.tradingday.Period import check_period
from ultron.tradingday.Schedule import Schedule
from ultron.tradingday.Enums import BizDayConventions
from ultron.tradingday.Enums import DateGeneration
from ultron.tradingday.Enums import TimeUnits


def isBizDay(holidayCenter, ref):
    cal = Calendar(holidayCenter)
    ref = check_date(ref)
    return cal.isBizDay(ref)


def datesList(fromDate, toDate):
    fromDate = check_date(fromDate)
    toDate = check_date(toDate)
    return [Date.fromExcelSerialNumber(serial).toDateTime() for serial in
            range(fromDate.serialNumber, toDate.serialNumber + 1)]


def bizDatesList(holidayCenter, fromDate, toDate):
    cal = Calendar(holidayCenter)
    fromDate = check_date(fromDate)
    toDate = check_date(toDate)
    return [d.toDateTime() for d in cal.bizDatesList(fromDate, toDate)]


def holDatesList(holidayCenter, fromDate, toDate, includeWeekend=True):
    cal = Calendar(holidayCenter)
    fromDate = check_date(fromDate)
    toDate = check_date(toDate)
    return [d.toDateTime() for d in cal.holDatesList(fromDate, toDate, includeWeekend)]


def advanceDate(referenceDate, period):
    d = check_date(referenceDate) + period
    return d.toDateTime()


def adjustDateByCalendar(holidayCenter, referenceDate, convention=BizDayConventions.Following):
    cal = Calendar(holidayCenter)
    refer = check_date(referenceDate)
    return cal.adjustDate(refer, convention).toDateTime()


def advanceDateByCalendar(holidayCenter, referenceDate, period, convention=BizDayConventions.Following):
    cal = Calendar(holidayCenter)
    refer = check_date(referenceDate)
    period = check_period(period)
    return cal.advanceDate(refer, period, convention).toDateTime()


def nthWeekDay(nth, dayOfWeek, month, year):
    date = Date.nthWeekday(nth, dayOfWeek, month, year)
    return date.toDateTime()


def makeSchedule(firstDate,
                 endDate,
                 tenor,
                 calendar='NullCalendar',
                 dateRule=BizDayConventions.Following,
                 dateGenerationRule=DateGeneration.Forward):

    cal = Calendar(calendar)
    firstDate = check_date(firstDate)
    endDate = check_date(endDate)
    tenor = check_period(tenor)

    if tenor.units() == TimeUnits.BDays:
        schedule = []
        if dateGenerationRule == DateGeneration.Forward:
            d = cal.adjustDate(firstDate, dateRule)
            while d <= endDate:
                schedule.append(d)
                d = cal.advanceDate(d, tenor, dateRule)
        elif dateGenerationRule == DateGeneration.Backward:
            d = cal.adjustDate(endDate, dateRule)
            while d >= firstDate:
                schedule.append(d)
                d = cal.advanceDate(d, -tenor, dateRule)
            schedule = sorted(schedule)
    else:
        schedule = Schedule(firstDate, endDate, tenor, cal, convention=dateRule, dateGenerationRule=dateGenerationRule)
    return [d.toDateTime() for d in schedule]



__all__ = ["datesList",
           "bizDatesList",
           "holDatesList",
           "isBizDay",
           "advanceDate",
           "BizDayConventions",
           "DateGeneration",
           "adjustDateByCalendar",
           "advanceDateByCalendar",
           "nthWeekDay",
           "makeSchedule"]



# -*- coding: utf-8 -*-

from collections import defaultdict
class Singleton(type):
    """
    单例metaclass
    """

    def __init__(cls, name, bases, dict):
        super(Singleton, cls).__init__(name, bases, dict)
        cls._instance = defaultdict()

    def __call__(cls, *args, **kw):
        tag = kw.get('tag') or 'default'
        if cls._instance.get(tag) is None:
            cls._instance[tag] = super(Singleton, cls).__call__(*args, **kw)
        return cls._instance[tag]



from typing import Union
import cvxpy as cp
import numpy as np
from .base import _IOptimizer, require


class QOptimizer(_IOptimizer):

    def __init__(self,
                 cost: np.ndarray,
                 variance: np.ndarray,
                 penalty: float = 1.0,
                 cons_matrix: np.ndarray = None,
                 lower_bound: Union[float, np.ndarray] = None,
                 upper_bound: Union[float, np.ndarray] = None):
        super().__init__(cost, cons_matrix, lower_bound, upper_bound)
        self._variance = variance
        self._penalty = penalty

    def solve(self, solver: str = "ECOS"):
        x, constraints = self._prepare()
        prob = cp.Problem(
            cp.Minimize(x @ self._cost +
                        0.5 * self._penalty * cp.quad_form(x, self._variance)),
            constraints=constraints)
        prob.solve(solver=solver)
        return x.value, prob.value, prob.status


class DecomposedQOptimizer(_IOptimizer):

    def __init__(self,
                 cost: np.ndarray,
                 factor_var: np.ndarray,
                 factor_load: np.ndarray,
                 factor_special: np.ndarray,
                 penalty: float = 1.0,
                 cons_matrix: np.ndarray = None,
                 lower_bound: Union[float, np.ndarray] = None,
                 upper_bound: Union[float, np.ndarray] = None):
        super().__init__(cost, cons_matrix, lower_bound, upper_bound)
        self._factor_var = factor_var
        self._factor_load = factor_load
        self._factor_special = factor_special
        self._penalty = penalty

    def solve(self, solver: str = "ECOS"):
        x, constraints = self._prepare()
        risk = cp.sum_squares(cp.multiply(np.sqrt(self._factor_special), x)) \
               + cp.quad_form((x.T @ self._factor_load).T, self._factor_var)
        prob = cp.Problem(cp.Minimize(x @ self._cost +
                                      0.5 * self._penalty * risk),
                          constraints=constraints)
        prob.solve(solver=solver)
        return x.value, prob.value, prob.status


class TargetVarianceOptimizer(_IOptimizer):

    def __init__(self,
                 cost: np.ndarray,
                 variance_target: float,
                 factor_var: np.ndarray = None,
                 factor_load: np.ndarray = None,
                 factor_special: np.ndarray = None,
                 variance: np.ndarray = None,
                 cons_matrix: np.ndarray = None,
                 lower_bound: Union[float, np.ndarray] = None,
                 upper_bound: Union[float, np.ndarray] = None):
        super().__init__(cost, cons_matrix, lower_bound, upper_bound)
        require(factor_var is not None or variance is not None, ValueError,
                "factor var or total var should not all be empty")
        if factor_var is not None:
            self._factor_var = factor_var
            self._factor_load = factor_load
            self._factor_special = factor_special
            self._use_factor = True
        else:
            self._variance = variance
            self._use_factor = False
        require(variance_target >= 0, ValueError,
                "variance target can't be negative")
        self._var_target = variance_target

    def solve(self, solver: str = "ECOS"):
        x, constraints = self._prepare()
        if self._use_factor:
            risk = cp.sum_squares(cp.multiply(np.sqrt(self._factor_special), x)) \
                   + cp.quad_form((x.T @ self._factor_load).T, self._factor_var)
        else:
            risk = cp.quad_form(x, self._variance)
        constraints.append(risk <= self._var_target)
        prob = cp.Problem(cp.Minimize(x @ self._cost), constraints=constraints)
        prob.solve(solver=solver)
        return x.value, prob.value, prob.status


# -*- encoding:utf-8 -*-
"""
    黄金分割及比例分割示例模块
"""
from collections import namedtuple
import matplotlib.pyplot as plt
from ultron.ump.technical import exectue as tl_exectue


def create_golden(kl_pd):
    """
    只针对金融时间序列的收盘价格close序列，进行黄金分割及比例分割
    数值结果分析以及可视化
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param show: 是否可视化黄金分割及比例分割结果
    :return: 黄金分割及比例分割结果组成的namedtuple数值对象
    """
    kl_close = kl_pd.close
    if not hasattr(kl_pd, 'name'):
        # 金融时间序列中如果有异常的没有name信息的补上一个unknown
        kl_pd.name = 'unknown'
    # 计算视觉黄金分割
    gd_382, gd_500, gd_618 = tl_exectue.find_golden_point(
        kl_pd.index, kl_close)
    # 计算统计黄金分割
    gex_382, gex_500, gex_618 = tl_exectue.find_golden_point_ex(
        kl_pd.index, kl_close)

    # below above 382, 618确定，即382，618上下底
    below618, above618 = tl_exectue.below_above_gen(gd_618, gex_618)
    below382, above382 = tl_exectue.below_above_gen(gd_382, gex_382)

    # 再次通过比例序列percents和find_percent_point寻找对应比例的位置字典pts_dict
    percents = [0.20, 0.25, 0.30, 0.70, 0.80, 0.90, 0.95]
    pts_dict = tl_exectue.find_percent_point(percents, kl_close)

    # 0.20, 0.25, 0.30只找最低的，即底部只要最低的
    below200, _ = tl_exectue.below_above_gen(*pts_dict[0.20])
    below250, _ = tl_exectue.below_above_gen(*pts_dict[0.25])
    below300, _ = tl_exectue.below_above_gen(*pts_dict[0.30])

    # 0.70, 0.80, 0.90, 0.95只找最高的，即顶部只要最高的
    _, above700 = tl_exectue.below_above_gen(*pts_dict[0.70])
    _, above800 = tl_exectue.below_above_gen(*pts_dict[0.80])
    _, above900 = tl_exectue.below_above_gen(*pts_dict[0.90])
    _, above950 = tl_exectue.below_above_gen(*pts_dict[0.95])

    return namedtuple('golden', [
        'g382', 'gex382', 'g500', 'gex500', 'g618', 'gex618', 'above618',
        'below618', 'above382', 'below382', 'above950', 'above900', 'above800',
        'above700', 'below300', 'below250', 'below200'
    ])(gd_382, gex_382, gd_500, gex_500, gd_618, gex_618, above618, below618,
       above382, below382, above950, above900, above800, above700, below300,
       below250, below200)



import numpy as np


def quantile(x: np.ndarray, n_bins: int) -> np.ndarray:

    n = x.size
    sorter = x.argsort()
    inv = np.empty(n, dtype=int)
    inv[sorter] = np.arange(n, dtype=int)

    bin_size = float(n) / n_bins

    pillars = [int(i * bin_size) for i in range(1, n_bins+1)]

    q_groups = np.empty(n, dtype=int)

    starter = 0
    for i, r in enumerate(pillars):
        q_groups[(inv >= starter) & (inv < r)] = i
        starter = r

    return q_groups


# -*- coding: utf-8 -*-
"""
Created on 2018-5-29

@author: cheng.li
"""

import abc
from typing import List

import pandas as pd


class RiskModel(metaclass=abc.ABCMeta):

    def get_risk_profile(self):
        pass


class FullRiskModel(RiskModel):

    def __init__(self, sec_cov: pd.DataFrame):
        self.codes = sec_cov.index.tolist()
        self.sec_cov = sec_cov.loc[self.codes, self.codes]

    def get_cov(self, codes: List[int] = None):
        if codes:
            return self.sec_cov.loc[codes, codes].values
        else:
            return self.sec_cov.values

    def get_risk_profile(self, codes: List[int] = None):
        return dict(cov=self.get_cov(codes),
                    factor_cov=None,
                    factor_loading=None,
                    idsync=None)


class FactorRiskModel(RiskModel):

    def __init__(self, factor_cov: pd.DataFrame, risk_exp: pd.DataFrame,
                 idsync: pd.Series):
        self.factor_cov = factor_cov
        self.idsync = idsync
        self.codes = self.idsync.index.tolist()
        self.factor_names = sorted(self.factor_cov.index)
        self.risk_exp = risk_exp.loc[self.codes, self.factor_names]
        self.factor_cov = self.factor_cov.loc[self.factor_names,
                                              self.factor_names]
        self.idsync = self.idsync[self.codes]

    def get_risk_exp(self, codes: List[int] = None):
        if codes:
            return self.risk_exp.loc[codes, :].values
        else:
            return self.risk_exp.values

    def get_factor_cov(self):
        return self.factor_cov.values

    def get_idsync(self, codes: List[int] = None):
        if codes:
            return self.idsync[codes].values
        else:
            return self.idsync.values

    def get_risk_profile(self, codes: List[int] = None):
        return dict(cov=None,
                    factor_cov=self.get_factor_cov(),
                    factor_loading=self.get_risk_exp(codes),
                    idsync=self.get_idsync(codes))



# -*- encoding:utf-8 -*-
"""
    买入择时示例因子：日胜率均值买入择时因子
"""

from ultron.ump.factor.buy.base import BuyCallMixin, FactorBuyTD
from ultron.ump.technical.line import Line


class FactorBuyXD(FactorBuyTD, BuyCallMixin):
    """
        示例短线：日胜率均值回复策略

        1. 默认以40天为周期(8周)结合涨跌阀值计算周几适合买入
        2. 回测运行中每一月重新计算一次上述的周几适合买入
        3. 在策略日任务中买入信号为：昨天下跌，今天开盘也下跌，且明天是计算出来的上涨概率大的'周几'
    """

    def _init_self(self, **kwargs):
        """
            kwargs中可选参数：buy_dw:    代表周期胜率阀值，默认0.55即55%
            kwargs中可选参数：buy_dwm:   代表涨幅比例阀值系数，默认0.618
            kwargs中可选参数：dw_period: 代表分析dw，dwm所使用的交易周期，默认40天周期(8周)
        """
        self.buy_dw = kwargs.pop('buy_dw', 0.55)
        self.buy_dwm = kwargs.pop('buy_dwm', 0.618)
        self.dw_period = kwargs.pop('dw_period', 40)

        # combine_kl_pd中包含择时金融时间数据与择时之前周期的金融时间数据, 先取出择时开始之前的周期数据
        last_kl = self.combine_kl_pd.loc[:self.kl_pd.index[0]]
        if last_kl.shape[0] > self.dw_period:
            last_kl = last_kl[-self.dw_period:]
        # 开始计算周几买，_make_buy_date把结果被放在self.buy_date_week序列中
        self._make_buy_date(last_kl)

    def fit_day(self, today):
        """日任务：昨天下跌，今天开盘也下跌，根据今天是周几，在不在序列self.buy_date_week中决定今天买不买"""
        if self.yesterday.p_change < 0 and today.open < self.yesterday.close \
                and int(today.date_week) in self.buy_date_week:
            # 由于没有用到今天的收盘价格等，可以直接使用buy_today
            return self.buy_today()
        return None

    def _make_buy_date(self, last_kl):
        """
        根据金融时间周期数据切片具体计算'周几'买上涨概率大

        :param last_kl: 金融时间周期数据切片，pd.DataFrame
        """
        from ultron.kdutils import kline as KLUtil

        self.buy_date_week = []
        # 计算周期内，周期的胜率
        last_dw = KLUtil.date_week_win(last_kl)
        # 摘取大于阀值self.buy_dw的'周几'，buy_dw默认0.55
        last_dw_vd = last_dw[last_dw.win >= self.buy_dw]
        """
            eg: last_dw_vd
                       0  1   win
            date_week
            周四         3  5  0.62
            周五         2  6  0.75
        """
        if len(last_dw_vd) > 0:
            # 如果胜率有符合要求的，使用周几平均涨幅计算date_week_mean
            last_dwm = KLUtil.date_week_mean(last_kl)
            # 摘取满足胜率的last_dw_vd
            last_dwm_vd = last_dwm.loc[last_dw_vd.index]
            """
                eg: last_dwm_vd
                           _p_change
                date_week
                周四              1.55
                周五              1.12
            """
            # 阀值计算方式1
            dwm1 = abs(last_dwm.sum()).values[0] / self.buy_dwm
            # 阀值计算方式2
            dwm2 = abs(last_dwm._p_change).mean() / self.buy_dwm
            # 如果symbol多可以使用&的关系
            dm_effect = (last_dwm_vd._p_change > dwm1) | (last_dwm_vd._p_change
                                                          > dwm2)
            buy_date_loc = last_dwm_vd[dm_effect].index
            """
                eg: buy_date_loc
                Index(['周四', '周五'], dtype='object', name='date_week')
            """
            if len(buy_date_loc) > 0:
                # 如果涨跌幅阀值也满足，tolist，eg：['周一', '周二', '周三', '周四', '周五']
                dw_index = last_dw.index.tolist()
                # 如果是一周5个交易日的就是4
                max_ind = len(dw_index) - 1
                for bdl in buy_date_loc:
                    sell_ind = dw_index.index(bdl)
                    buy_ind = sell_ind - 1 if sell_ind > 0 else max_ind
                    self.buy_date_week.append(buy_ind)


# -*- encoding:utf-8 -*-
"""
    使用的金融时间序列分析模块, 模块真的方法真的参数都为格式化好的kl如下

    eg:
                close	high	low	p_change	open	pre_close	volume	date	date_week	key	atr21	atr14
    2016-07-20	228.36	229.800	225.00	1.38	226.47	225.26	2568498	20160720	2	499	9.1923	8.7234
    2016-07-21	220.50	227.847	219.10	-3.44	226.00	228.36	4428651	20160721	3	500	9.1711	8.7251
    2016-07-22	222.27	224.500	218.88	0.80	221.99	220.50	2579692	20160722	4	501	9.1858	8.7790
    2016-07-25	230.01	231.390	221.37	3.48	222.27	222.27	4490683	20160725	0	502	9.2669	8.9298
    2016-07-26	225.93	228.740	225.63	-1.77	227.34	230.01	41833	20160726	1	503	9.1337	8.7541
"""
import pandas as pd
import numpy as np
from collections import Iterable
from ultron.utilities.logger import kd_logger


def _df_dispatch(df, dispatch_func):
    """
    根据df的类型分发callable的执行方法，

    :param df: 格式化好的kl，或者字典，或者可迭代序列
    :param dispatch_func: 分发的可执行的方法
    """
    if isinstance(df, pd.DataFrame):
        # 参数只是pd.DataFrame
        return dispatch_func(df)
    elif isinstance(df, dict) and all(
        [isinstance(_df, pd.DataFrame) for _df in df.values()]):
        # 参数只是字典形式
        return [dispatch_func(df[df_key], df_key) for df_key in df]
    elif isinstance(df, Iterable) and all(
        [isinstance(_df, pd.DataFrame) for _df in df]):
        # 参数只是可迭代序列
        return [dispatch_func(_df) for _df in df]
    else:
        kd_logger.error('df type is error! {}'.format(type(df)))


def _df_dispatch_concat(df, dispatch_func):
    """
    根据df的类型分发callable的执行方法，如果是字典或者可迭代类型的返回值使用
    pd.concat连接起来

    :param df: 格式化好的kl，或者字典，或者可迭代序列
    :param dispatch_func: 分发的可执行的方法
    """

    if isinstance(df, pd.DataFrame):
        # 参数只是pd.DataFrame
        return dispatch_func(df)
    elif isinstance(df, dict) and all(
        [isinstance(_df, pd.DataFrame) for _df in df.values()]):
        # 参数只是字典形式
        return pd.concat([dispatch_func(df[df_key], df_key) for df_key in df],
                         axis=1)
    elif isinstance(df, Iterable) and all(
        [isinstance(_df, pd.DataFrame) for _df in df]):
        # 参数只是可迭代序列
        return pd.concat([dispatch_func(_df) for _df in df], axis=1)
    else:
        kd_logger.error('df type is error! {}'.format(type(df)))


def wave_change_rate(df):
    """
    eg:
        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.wave_change_rate(tsla)

        out:
        日振幅涨跌幅比：1.794156

    :param df: 格式化好的kl，或者字典，或者可迭代序列
    """

    def _wave_change_rate(p_df, df_name=''):
        wave = ((p_df.high - p_df.low) / p_df.pre_close) * 100
        # noinspection PyUnresolvedReferences
        wave_rate = wave.mean() / np.abs(p_df['p_change']).mean()
        return {'code': df_name, 'wave_rate': wave_rate}  # 1.80
        #print('{}日振幅涨跌幅比：{:2f}, {}日统计套利条件'.format(
        #    df_name, wave_rate, '具备' if wave_rate > 1.80 else '不具备'))

    return _df_dispatch(df, _wave_change_rate)


def p_change_stats(df):
    """
    eg :
        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.p_change_stats(tsla)

        out:

        日涨幅平均值1.861, 共260个交易日上涨走势
        日跌幅平均值-1.906, 共244个交易日下跌走势
        日平均涨跌比0.977, 上涨下跌数量比:1.066

    :param df: 格式化好的kl，或者字典，或者可迭代序列
    """

    def _p_change_stats(p_df, df_name=''):
        p_change_up = p_df[p_df['p_change'] > 0].p_change
        p_change_down = p_df[p_df['p_change'] < 0].p_change
        return {
            'code': df_name,
            'up_mean': p_change_up.mean(),
            'up_count': p_change_up.count(),
            'down_mean': p_change_down.mean(),
            'down_count': p_change_down.count(),
            'chg_mean': abs(p_change_up.mean() / p_change_down.mean()),
            'chg_ratio': p_change_up.count() / p_change_down.count()
        }

    return _df_dispatch(df, _p_change_stats)


def date_week_wave(df):
    """
    根据周几分析金融时间序列中的日波动:

    eg:
        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.date_week_wave(tsla)

        out:
            usTSLAwave
            date_week
            周一  3.8144
            周二  3.3326
            周三  3.3932
            周四  3.3801
            周五  2.9923

    :param df: ultronpy中格式化好的kl，或者字典，或者可迭代序列
    :return: pd.Series或者pd.DataFrame
    """

    def _date_week_wave(p_df, df_name=''):
        # 要改df所以copy
        df_copy = p_df.copy()
        wave_key = '{}wave'.format(df_name)
        # 计算波动: * 100目的是和金融序列中的p_change单位一致
        df_copy[wave_key] = (
            (df_copy.high - df_copy.low) / df_copy.pre_close) * 100
        dww = df_copy.groupby('date_week')[wave_key].mean()
        return dww

    return _df_dispatch_concat(df, _date_week_wave)


def date_week_win(df):
    """
    eg:
        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.date_week_win(tsla)

        out：
                0	1	win
            date_week
            0	44	51	0.5368
            1	55	48	0.4660
            2	48	57	0.5429
            3	44	57	0.5644
            4	53	47	0.470

    :param df: bupy中格式化好的kl，或者字典，或者可迭代序列
    :return: pd.Series或者pd.DataFrame
    """

    def _date_week_win(p_df, df_name=''):
        _df = p_df.copy()
        win_key = '{}win'.format(df_name)
        _df[win_key] = _df['p_change'] > 0
        _df[win_key] = _df[win_key].astype(int)

        dww = pd.concat([
            pd.crosstab(_df.date_week, _df[win_key]),
            _df.pivot_table([win_key], index='date_week')
        ],
                        axis=1)
        # 将周几这个信息变成中文
        # noinspection PyUnresolvedReferences
        return dww

    return _df_dispatch_concat(df, _date_week_win)


def date_week_mean(df):
    """
        eg:

        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.date_week_mean(tsla)

        out:
        周一    0.0626
        周二    0.0475
        周三    0.0881
        周四    0.2691
        周五   -0.2838
    :param df: ultronpy中格式化好的kl，或者字典，或者可迭代序列
    :return: pd.Series或者pd.DataFrame
    """

    def _date_week_win(p_df, df_name=''):
        dww = p_df.groupby('date_week')['p_change'].mean()
        # p_change变成对应的pchange
        dww = pd.DataFrame(dww)
        dww.rename(columns={'p_change': '{}_p_change'.format(df_name)},
                   inplace=True)
        return dww

    return _df_dispatch_concat(df, _date_week_win)


def bcut_change_vc(df, bins=None):
    """
    eg:
        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.bcut_change_vc(tsla)

        out:
                p_change	rate
        (0, 3]	209	0.4147
        (-3, 0]	193	0.3829
        (3, 7]	47	0.0933
        (-7, -3]	44	0.0873
        (-10, -7]	6	0.0119
        (7, 10]	3	0.0060
        (10, inf]	1	0.0020
        (-inf, -10]	1	0.0020

    :param df: ultronpy中格式化好的kl，或者字典，或者可迭代序列
    :param bins: 默认eg：[-np.inf, -10, -7, -3, 0, 3, 7, 10, np.inf]
    :return: pd.DataFrame
    """

    def _bcut_change_vc(p_df, df_name=''):
        dww = pd.DataFrame(pd.cut(p_df.p_change, bins=bins).value_counts())
        # 计算各个bin所占的百分比
        dww['{}rate'.format(
            df_name)] = dww.p_change.values / dww.p_change.values.sum()
        if len(df_name) > 0:
            dww.rename(columns={'p_change': '{}'.format(df_name)},
                       inplace=True)
        return dww

    if bins is None:
        bins = [-np.inf, -10, -7, -3, 0, 3, 7, 10, np.inf]
    return _df_dispatch_concat(df, _bcut_change_vc)


def qcut_change_vc(df, q=10):
    """
    eg:
        tsla = SymbolPd.make_kl_df('usTSLA')
        KLUtil.qcut_change_vc(tsla)

        out:
            change
        0	[-10.45, -3.002]
        1	(-3.002, -1.666]
        2	(-1.666, -0.93]
        3	(-0.93, -0.396]
        4	(-0.396, 0.065]
        5	(0.065, 0.48]
        6	(0.48, 1.102]
        7	(1.102, 1.922]
        8	(1.922, 3.007]
        9	(3.007, 11.17]

    :param df: ultronpy中格式化好的kl，或者字典，或者可迭代序列
    :param q: 透传qcut使用的q参数，默认10，10等分
    :return: pd.DataFrame
    """

    def _qcut_change_vc(p_df, df_name=''):
        dww = pd.qcut(p_df.p_change, q).value_counts().index.values
        # 构造Categories使用DataFrame套Series
        dww = pd.Series(dww)
        # 涨跌从负向正开始排序
        dww.sort_values(inplace=True)
        dww = pd.DataFrame(dww)
        # 排序后index重新从0开始排列
        dww.index = np.arange(0, q)
        dww.columns = ['{}change'.format(df_name)]
        return dww

    return _df_dispatch_concat(df, _qcut_change_vc)


# -*- encoding:utf-8 -*-
"""
    KMEANS 聚类寻优
"""
import numpy as np
import pandas as pd
from six.moves import xrange
from sklearn import metrics
from sklearn.cluster import KMeans as KmeansImpl
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress


class KMeans(object):

    def __init__(self,
                 min=0,
                 max=0,
                 method='squares',
                 features=None,
                 thresh=1):
        if method == 'squares':
            self._k_rng = xrange(min, max) if min != 0 else xrange(
                1, len(features))
        else:
            self._k_rng = xrange(min, max) if min != 0 else xrange(
                2, len(features))
        self._method = method
        self._features = features
        self._thresh = thresh

    def _silhoette_score(self, factors_data, est_arr):
        score = []
        i = 0
        with Progress(len(est_arr), 0, 'evaluate silhoette') as pg:
            for e in est_arr:
                i += 1
                pg.show(i)
                score.append(
                    metrics.silhouette_score(factors_data,
                                             e.labels_,
                                             metric='euclidean'))
        score = pd.Series(score, name='silhouette')
        score = score.sort_values()
        return score

    def _calinski_harabasz_score(self, factors_data, est_arr):
        score = []
        i = 0
        with Progress(len(est_arr), 0, 'evaluate harabasz') as pg:
            for e in est_arr:
                i += 1
                pg.show(i)
                score.append(
                    metrics.calinski_harabasz_score(factors_data, e.labels_))
        score = pd.Series(score, name='harabasz')
        score = score.sort_values()
        return score

    def _squares_score(self, est_arr):
        sum_squares = []
        i = 0
        with Progress(len(est_arr), 0, 'evaluate squares') as pg:
            for e in est_arr:
                i += 1
                pg.show(i)
                sum_squares.append(e.inertia_)
        diff_squares = [squares / sum_squares[0] for squares in sum_squares]
        diff_squares_pd = pd.Series(diff_squares, name='squares')
        score = diff_squares_pd[diff_squares_pd < self._thresh]
        score = score.sort_values()
        return score

    def _kmean_evaluate(self, factors_data, est_arr, method):
        if method == 'silhoette':
            score = self._silhoette_score(factors_data=factors_data,
                                          est_arr=est_arr)
        elif method == 'harabasz':
            score = self._calinski_harabasz_score(factors_data=factors_data,
                                                  est_arr=est_arr)
        elif method == 'squares':
            score = self._squares_score(est_arr=est_arr)

        select_k = self._k_rng[score.index[0]]
        return select_k

    def fit_kmeans(self, factors_data):
        est_arr = []
        i = 0
        with Progress(len(self._k_rng), 0, 'fit kmeans') as pg:
            for k in self._k_rng:
                i += 1
                est_arr.append(KmeansImpl(n_clusters=k).fit(factors_data))
                pg.show(i)
        return est_arr

    def run(self, factors_data):
        diff_cols = [
            col for col in factors_data.columns if col not in self._features
        ]
        kd_logger.info("initialize kmeans range:{0}".format(self._k_rng))
        est_arr = self.fit_kmeans(factors_data[self._features])
        kd_logger.info("{} evaluate kmean best k".format(self._method))
        best_k = self._kmean_evaluate(
            factors_data=factors_data[self._features],
            est_arr=est_arr,
            method=self._method)

        kd_logger.info("kmean best {}".format(best_k))

        est = KmeansImpl(n_clusters=best_k).fit(factors_data[self._features])

        kmeans_clustering_labels = pd.DataFrame(est.labels_,
                                                columns=['cluster'])
        kmeans_data = pd.concat([
            factors_data[diff_cols + self._features], kmeans_clustering_labels
        ],
                                axis=1)
        return kmeans_data


# -*- coding: utf-8 -*-


import copy
import six
import numpy as np
cimport numpy as np
cimport cython
from ultron.sentry.Analysis.SeriesValues cimport SeriesValues
from ultron.sentry.Analysis.SecurityValueHolders cimport SecurityValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import build_holder
from ultron.sentry.Math.MathConstants cimport NAN


cdef class CrossSectionValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _inner
    cdef public SecurityValueHolder _group

    def __init__(self, innerValue, groups=None):
        super(CrossSectionValueHolder, self).__init__()
        self._inner = build_holder(innerValue)
        self._group = build_holder(groups) if groups else None
        if self._group:
            self._window = max(self._inner.window, self._group.window)
            self._dependency = list(set(self._inner.fields + self._group.fields))
        else:
            self._window = self._inner.window
            self._dependency = copy.deepcopy(self._inner.fields)
        self.updated = 0
        self.cached = None

    @property
    def symbolList(self):
        return self._inner.symbolList

    cpdef push(self, dict data):
        self._inner.push(data)
        if self._group:
            self._group.push(data)
        self.updated = 0


cdef class CSTopNSecurityValueHolder(CrossSectionValueHolder):

    cdef int _n;

    def __init__(self, innerValue, n, groups=None):
        super(CSTopNSecurityValueHolder, self).__init__(innerValue, groups)
        self._n = n

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.top_n(self._n, self._group.value_all())
        else:
            self.cached = raw_values.top_n(self._n)
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.top_n(self._n, self._group.value_by_names(names))
        else:
            raw_values = raw_values.top_n(self._n)
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSTopN}}({0}, {1}, groups={2})".format(str(self._inner), self._n, str(self._group))
        else:
            return "\mathrm{{CSTopN}}({0}, {1})".format(str(self._inner), self._n)


cdef class CSTopNPercentileSecurityValueHolder(CrossSectionValueHolder):

    cdef double _n;

    def __init__(self, innerValue, n, groups=None):
        super(CSTopNPercentileSecurityValueHolder, self).__init__(innerValue, groups)
        self._n = n

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.top_n_percentile(self._n, self._group.value_all())
        else:
            self.cached = raw_values.top_n_percentile(self._n)
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.top_n_percentile(self._n, self._group.value_by_names(names))
        else:
            raw_values = raw_values.top_n_percentile(self._n)
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSTopNPercentile}}({0}, {1}, groups={2})".format(str(self._inner), self._n, str(self._group))
        else:
            return "\mathrm{{CSTopNPercentile}}({0}, {1})".format(str(self._inner), self._n)


cdef class CSBottomNSecurityValueHolder(CrossSectionValueHolder):

    cdef int _n;

    def __init__(self, innerValue, n, groups=None):
        super(CSBottomNSecurityValueHolder, self).__init__(innerValue, groups)
        self._n = n

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.bottom_n(self._n, self._group.value_all())
        else:
            self.cached = raw_values.bottom_n(self._n)
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.bottom_n(self._n, self._group.value_by_names(names))
        else:
            raw_values = raw_values.bottom_n(self._n)
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSBottomN}}({0}, {1}, groups={2})".format(str(self._inner), self._n, str(self._group))
        else:
            return "\mathrm{{CSBottomN}}({0}, {1})".format(str(self._inner), self._n)


cdef class CSBottomNPercentileSecurityValueHolder(CrossSectionValueHolder):

    cdef double _n;

    def __init__(self, innerValue, n, groups=None):
        super(CSBottomNPercentileSecurityValueHolder, self).__init__(innerValue, groups)
        self._n = n

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.bottom_n_percentile(self._n, self._group.value_all())
        else:
            self.cached = raw_values.bottom_n_percentile(self._n)
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.bottom_n_percentile(self._n, self._group.value_by_names(names))
        else:
            raw_values = raw_values.bottom_n_percentile(self._n)
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSBottomNPercentile}}({0}, {1}, groups={2})".format(str(self._inner), self._n, str(self._group))
        else:
            return "\mathrm{{CSBottomNPercentile}}({0}, {1})".format(str(self._inner), self._n)


cdef class CSRankedSecurityValueHolder(CrossSectionValueHolder):

    def __init__(self, innerValue, groups=None):
        super(CSRankedSecurityValueHolder, self).__init__(innerValue, groups)

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.rank(self._group.value_all())
        else:
            self.cached = raw_values.rank()
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.rank(self._group.value_by_names(names))
        else:
            raw_values = raw_values.rank()
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSRank}}({0}, groups={1})".format(str(self._inner), str(self._group))
        else:
            return "\mathrm{{CSRank}}({0})".format(str(self._inner))


cdef class CSAverageSecurityValueHolder(CrossSectionValueHolder):
    def __init__(self, innerValue, groups=None):
        super(CSAverageSecurityValueHolder, self).__init__(innerValue, groups)

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.mean(self._group.value_all())
        else:
            self.cached = raw_values.mean()
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.mean(self._group.value_by_names(names))
        else:
            raw_values = raw_values.mean()
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSMean}}({0}, groups={1})".format(str(self._inner), str(self._group))
        else:
            return "\mathrm{{CSMean}}({0})".format(str(self._inner))


cdef class CSPercentileSecurityValueHolder(CrossSectionValueHolder):

    def __init__(self, innerValue, groups=None):
        super(CSPercentileSecurityValueHolder, self).__init__(innerValue, groups)

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.percentile(self._group.value_all())
        else:
            self.cached = raw_values.percentile()
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.percentile(self._group.value_by_names(names))
        else:
            raw_values = raw_values.percentile()
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSPercentile}}({0}, groups={1})".format(str(self._inner), str(self._group))
        else:
            return "\mathrm{{CSPercentile}}({0})".format(str(self._inner))


cdef class CSAverageAdjustedSecurityValueHolder(CrossSectionValueHolder):
    def __init__(self, innerValue, groups=None):
        super(CSAverageAdjustedSecurityValueHolder, self).__init__(innerValue, groups)

    cpdef value_all(self):

        cdef SeriesValues raw_values

        if self.updated:
            return self.cached
        else:
            raw_values = self._inner.value_all()
            if self._group:
                self.cached = raw_values - raw_values.mean(self._group.value_all())
            else:
                self.cached = raw_values - raw_values.mean()
            self.updated = 1
            return self.cached

    cpdef double value_by_name(self, name):

        cdef SeriesValues raw_values

        if self.updated:
            return self.cached[name]
        else:
            raw_values = self._inner.value_all()
            if self._group:
                self.cached = raw_values - raw_values.mean(self._group.value_all())
            else:
                self.cached = raw_values - raw_values.mean()
            self.updated = 1
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        raw_values = self._inner.value_by_names(names)
        if self._group:
            self.cached = raw_values - raw_values.mean(self._group.value)
        else:
            self.cached = raw_values - raw_values.mean()
        return raw_values[names]

    def __str__(self):
        if self._group:
            return "\mathrm{{CSMeanAdjusted}}({0}, groups={1})".format(str(self._inner), str(self._group))
        else:
            return "\mathrm{{CSMeanAdjusted}}({0})".format(str(self._inner))


cdef class CSZScoreSecurityValueHolder(CrossSectionValueHolder):
    def __init__(self, innerValue, groups=None):
        super(CSZScoreSecurityValueHolder, self).__init__(innerValue, groups)

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.zscore(self._group.value_all())
        else:
            self.cached = raw_values.zscore()
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.zscore(self._group.value_by_names(names))
        else:
            raw_values = raw_values.zscore()
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSZscore}}({0}, groups={1})".format(str(self._inner), str(self._group))
        else:
            return "\mathrm{{CSZscore}}({0})".format(str(self._inner))


cdef class CSFillNASecurityValueHolder(CrossSectionValueHolder):
    def __init__(self, innerValue, groups=None):
        super(CSFillNASecurityValueHolder, self).__init__(innerValue, groups)

    cdef _cal_impl(self):
        cdef SeriesValues raw_values = self._inner.value_all()

        if self._group:
            self.cached = raw_values.fillna(self._group.value_all())
        else:
            self.cached = raw_values.fillna()
        self.updated = 1

    cpdef value_all(self):
        if self.updated:
            return self.cached
        else:
            self._cal_impl()
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            self._cal_impl()
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues raw_values = self._inner.value_by_names(names)
        if self._group:
            raw_values = raw_values.fillna(self._group.value_by_names(names))
        else:
            raw_values = raw_values.fillna()
        return raw_values

    def __str__(self):
        if self._group:
            return "\mathrm{{CSFillNA}}({0}, groups={1})".format(str(self._inner), str(self._group))
        else:
            return "\mathrm{{CSFillNA}}({0})".format(str(self._inner))


cdef class CSResidueSecurityValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _left
    cdef public SecurityValueHolder _right

    def __init__(self, left, right):
        super(CSResidueSecurityValueHolder, self).__init__()
        self._left = build_holder(left)
        self._right = build_holder(right)

        self._window = max(self._left.window, self._right.window)
        self._dependency = list(set(self._left.fields + self._right.fields))
        self.updated = 0
        self.cached = None

    @property
    def symbolList(self):
        return self._left.symbolList

    cpdef push(self, dict data):
        self._left.push(data)
        self._right.push(data)
        self.updated = 0

    cpdef value_all(self):

        cdef SeriesValues left_raw_values
        cdef SeriesValues right_raw_values

        if self.updated:
            return self.cached
        else:
            left_raw_values = self._left.value_all()
            right_raw_values = self._right.value_all()
            self.cached = left_raw_values.res(right_raw_values)
            self.updated = 1
            return self.cached

    cpdef double value_by_name(self, name):

        cdef SeriesValues left_raw_values
        cdef SeriesValues right_raw_values

        if self.updated:
            return self.cached[name]
        else:
            left_raw_values = self._left.value_all()
            right_raw_values = self._right.value_all()
            self.cached = left_raw_values.res(right_raw_values)
            self.updated = 1
            return self.cached[name]

    cpdef SeriesValues value_by_names(self, list names):
        cdef SeriesValues left_raw_values
        cdef SeriesValues right_raw_values
        left_raw_values = self._left.value_by_names(names)
        right_raw_values = self._right.value_by_names(names)
        raw_values = left_raw_values.res(right_raw_values)
        return raw_values

    def __str__(self):
        return "\mathrm{{CSRes}}({0}, {1})".format(str(self._left), str(self._right))



from celery import Celery
from celery.result import AsyncResult
from celery.schedules import crontab
import six
from ultron.utilities.singleton import Singleton
from ultron.config import config_setting


@six.add_metaclass(Singleton)
class AppEngine(object):

    def __init__(self, **kwargs):
        if ('host' in kwargs) and ('port' in kwargs) and ('pwd' in kwargs):
            self._host = kwargs['host']
            self._port = kwargs['port']
            if 'user' in kwargs:
                self._user = kwargs['user']
            self._pwd = kwargs['pwd']
            self._db = kwargs['db']
        else:
            self._host = config_setting.queue_host
            self._port = config_setting.queue_port
            self._user = config_setting.queue_user
            self._pwd = config_setting.queue_pwd
            self._db = config_setting.queue_db

    def create_engine(self, task_name, module_list=[]):
        #redis_url = 'redis://:12345678dx@127.0.0.1:6379/0'
        amqp_url = 'amqp://' + self._user + ':' + self._pwd + '@' + self._host + ':' + str(
            self._port) + '/' + str(self._db)
        app = Celery(task_name, broker=amqp_url, backend=amqp_url)

        class Config:
            CELERY_TIMEZONE = 'Asia/Shanghai'
            if len(module_list) > 0:
                CELERY_IMPORTS = tuple(module_list)

        app.config_from_object(Config)
        return app


create_app = AppEngine().create_engine



# -*- coding: utf-8 -*-
import unittest

import pandas as pd
import numpy as np

from ultron.optimize.model.linearmodel import LinearRegression
from ultron.optimize.model.loader import load_model

class Loader(unittest.TestCase):
    def setUp(self):
        self.n = 3
        self.trained_x = pd.DataFrame(np.random.randn(1000, self.n), columns=['a', 'b', 'c'])
        self.trained_y = np.random.randn(1000, 1)

        self.predict_x = pd.DataFrame(np.random.randn(100, self.n), columns=['a', 'b', 'c'])
    
    def test_load_model(self):
        model = LinearRegression(['a','b','c'])
        model.fit(self.trained_x, self.trained_y)

        model_desc = model.save()
        new_model = load_model(model_desc)

        np.testing.assert_array_almost_equal(model.predict(self.predict_x),
                                             new_model.predict(self.predict_x))

        self.assertEqual(model.features, new_model.features)

        self.assertEqual(model.trained_time, new_model.trained_time)



# -*- encoding:utf-8 -*-
"""
    卖出择时示例因子：n日卖出策略，不管什么结果，买入后只持有N天
"""

from ultron.ump.factor.sell.base import FactorSellBase, ESupportDirection


class FactorSellNDay(FactorSellBase):
    """n日卖出策略，不管交易现在什么结果，买入后只持有N天"""

    def _init_self(self, **kwargs):
        """kwargs中可以包含: 参数sell_n：代表买入后持有的天数，默认1天"""
        self.sell_n = kwargs.pop('sell_n', 1)
        self.is_sell_today = kwargs.pop('is_sell_today', False)
        self.sell_type_extra = '{}:sell_n={}'.format(self.__class__.__name__,
                                                     self.sell_n)

    def support_direction(self):
        """因子支持两个方向"""
        return [
            ESupportDirection.DIRECTION_CAll.value,
            ESupportDirection.DIRECTION_PUT.value
        ]

    def fit_day(self, today, orders):
        """
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        :return:
        """
        for order in orders:
            # 将单子的持有天数进行增加
            order.keep_days += 1
            if order.keep_days >= self.sell_n:
                # 只要超过self.sell_n即卖出
                self.sell_today(
                    order) if self.is_sell_today else self.sell_tomorrow(order)



# -*- encoding:utf-8 -*-
"""
    股票类型的symbol模块，a股，美股，港股
"""
import os
import pandas as pd
from ultron.env import g_base_path
from ultron.ump.core.env import EMarketSubType
from ultron.ump.core.base import FreezeAttrMixin
from ultron.ump.core.fixes import six
from ultron.kdutils.decorator import singleton
from ultron.kdutils.str_util import digit_str
"""a股symbol，文件定期重新爬取，更新"""
_stock_code_cn = os.path.join(g_base_path, 'stock_code_CN.csv')
_stock_code_us = os.path.join(g_base_path, 'stock_code_US.csv')


class StockBaseWrap(object):
    """做为类装饰器封装替换init 解析csv symbol数据操作，装饰替换init"""

    def __call__(self, cls):
        if isinstance(cls, six.class_types):
            # 只做为类装饰器使用，拿出原始的__init__
            init = cls.__init__

            def wrapped(*args, **kwargs):
                warp_self = args[0]
                warp_self.df = None

                init(*args, **kwargs)

                # 剔除重复数据
                warp_self.df.drop_duplicates(inplace=True)

                warp_self.df['industry_factorize'], industry_factorize_name = \
                    pd.factorize(warp_self.df.industry)
                # 用Series包装一下离散后的行业信息，以便方便对应行业索引
                warp_self.industry_factorize_name_series = pd.Series(
                    industry_factorize_name)

                # 将映射中key和value进行互换，columns_map中中文的key和英文的value（详ABuXqConsts），即形成本地语言列名
                #local_columns_map = {
                #    columns_map[col_key]: col_key
                #    for col_key in columns_map
                #}
                local_columns_map = {}
                # show_df中列名使用本地语言展示
                warp_self.show_df = warp_self.df.rename(
                    columns=local_columns_map, inplace=False)
                # 冻结接口，只读
                # noinspection PyProtectedMember
                warp_self._freeze()

            # 使用wrapped替换初始的__init__
            cls.__init__ = wrapped
            wrapped.__name__ = '__init__'
            # 记录原始init的方法为deprecated_original
            wrapped.deprecated_original = init
            return cls
        else:
            raise TypeError('StockBaseWrap just for class warp')


# noinspection PyUnresolvedReferences
class SymbolStockBase(FreezeAttrMixin):
    """股票类型的symbol抽象基类"""

    def __str__(self):
        """打印对象显示：df.info， df.describe"""
        return 'info:\n{}\ndescribe:\n{}'.format(self.df.info(),
                                                 self.df.describe())

    __repr__ = __str__

    def __len__(self):
        """对象长度：df.shape[0]，即行数"""
        return self.df.shape[0]

    def __setitem__(self, key, value):
        """索引设置：对外抛出错误， 即不准许外部设置"""
        raise AttributeError("AbuFuturesCn set value!!!")

    def query_industry_symbols(self, query_symbol, local_df=True):
        """
        为ABuIndustries模块，提供查询股票所在的行业industry_df子集
        :param query_symbol: symbol str对象
        :param local_df: 是否基于show_df返回行业df
        :return: 查询query_symbol所在的行业对象，pd.DataFrame对象
        """
        industry_df = None
        # 使用in查询self，即子类都需要实现自己的__contains__
        if query_symbol in self:
            # 忽略一个问题，如果只使用000001不带子市场标识去查询，结果只取第一个，准确查询需要完整标示
            factorize = self[query_symbol]['industry_factorize'].values[0]
            # 参数local_df决定行业industry_df子集基于本地语言的show_df还是df，使用show_df更加方便外面查阅对应内容意义
            base_df = self.show_df if local_df else self.df
            # 切取行业factorize相同的子集
            industry_df = base_df[base_df['industry_factorize'] == factorize]
        return industry_df

    def query_industry_factorize(self, factorize, local_df=True):
        """
        为ABuIndustries模块，提行业的factorize值查询industry_df子集
        :param factorize: int
        :param local_df: 是否基于show_df返回行业df
        :return: 通过factorize值查询industry_df子集，pd.DataFrame对象
        """
        base_df = self.show_df if local_df else self.df
        # 切取行业factorize相同的子集
        industry_df = base_df[base_df['industry_factorize'] == factorize]
        return industry_df

    def all_symbol(self, index=False):
        """
        子类需要实现，获取市场中所有股票symbol str对象序列，即全市场symbol序列
        :param index: 是否需要返回大盘symbol
        :return: 全市场symbol序列
        """
        raise NotImplementedError(
            'NotImplementedError AbuSymbolStockBase all_symbol!!!')

    def query_symbol_sub_market(self, *args, **kwargs):
        """
        子类需要实现，查询股票所在的子市场，即交易所信息
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """
        raise NotImplementedError(
            'NotImplementedError AbuSymbolStockBase query_symbol_sub_market!!!'
        )

    def symbol_func(self, df):
        """
        子类需要实现，通过df组装支持ABuSymbolPd.make_kl_df使用的symbol
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        raise NotImplementedError(
            'NotImplementedError AbuSymbolStockBase symbol_func!!!')


@singleton
@StockBaseWrap()
class SymbolCN(SymbolStockBase):
    """a股symbol类，singleton"""

    def __init__(self):
        """被AbuStockBaseWrap替换__init__，即只需读取a股数据到self.df 后续在类装饰器完成"""
        self.df = pd.read_csv(_stock_code_cn, index_col=0, dtype=str)

    def __contains__(self, item):
        """成员测试：是否item在self.df.symbol.values中"""
        return digit_str(item) in self.df.symbol.values

    def __getitem__(self, key):
        """
            索引获取：两种模式索引获取：
            1. 参数key为df的columns名称，返回self.df[key]，即get df的列
            2. 参数key为股票代码名称，标准化后查询，self.df[self.df.symbol == key]，即get df的行
        """

        if key in self.df.columns:
            # 参数key为df的columns名称，get df的行
            return self.df[key]

        if len(key) > 2:
            head = key[:2].upper()
            if head.isalpha():
                # 头两位是字面，即认为是exchange信息，直接截取df_filter
                df_filter = self.df[self.df['exchange'] == head]
                if not df_filter.empty:
                    if key[2:] in df_filter.symbol.values:
                        # get df的行信息，即对应股票的所有信息
                        return df_filter[df_filter.symbol == key[2:]]
            else:
                if key in self.df.symbol.values:
                    # get df的行
                    return self.df[self.df.symbol == key]

    def symbol_func(self, df):
        """
        通过df组装支持ABuSymbolPd.make_kl_df使用的symbol，
        使用df['exchange'].map(lambda exchange: exchange.lower()) + df['symbol']
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        df_symbol = df['exchange'].map(
            lambda exchange: exchange.lower()) + df['symbol']
        return df_symbol.tolist()

    def all_symbol(self, index=False):
        """
        获取a股市场中所有股票symbol str对象序列，即a股全市场symbol序列
        :param index: 是否需要返回a股大盘symbol
        :return: a股全市场symbol序列
        """

        # 过滤df中的A股指数symbol
        a_index = self.industry_factorize_name_series[
            self.industry_factorize_name_series == 'A股指数'].index.values[0]
        df_filter = self.df[self.df['industry_factorize'] != a_index]
        # 通过symbol_func转换为外部可直接使用ABuSymbolPd.make_kl_df请求的symbol序列
        all_symbol = self.symbol_func(df_filter)
        if index:
            # 需要返回大盘symbol
            all_symbol.extend([
                '{}{}'.format(EMarketSubType.SH.value, symbol)
                for symbol in Symbol.SH_INDEX
            ])
            all_symbol.extend([
                '{}{}'.format(EMarketSubType.SZ.value, symbol)
                for symbol in Symbol.SZ_INDEX
            ])
        return all_symbol

    def query_symbol_sub_market(self, code, default=EMarketSubType.SH.value):
        """
        查询股票所在的子市场，即交易所信息, A股市场默认返回上证交易所
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """

        if code in self:
            # 忽略一个问题，如果只使用000001不带子市场标识去查询，结果只取第一个，准确查询需要完整标示
            return self[code].market.values[0].lower()

        # 如果没查到如果首symbol为6，9为判定为sh
        if code[:1] in ['6', '9']:
            return EMarketSubType.SH.value
        # 如果没查到如果首symbol为2，3为判定为sz
        elif code[:1] in ['2', '3']:
            return EMarketSubType.SZ.value
        return default


@singleton
@StockBaseWrap()
class SymbolUS(SymbolStockBase):
    """美股symbol类，singleton"""
    """针对历史不适合做回测，对回测结果有误导影响的symbol, 即可能会产生几千倍，几百倍收益的不要参加回测"""
    s_unusual_symbol = [
        'usACV', 'usAMPH', 'usCBX', 'usDCIX', 'usDM', 'usEPE', 'usFPL',
        'usFUEL', 'usGDI', 'usHCC', 'usKBSF', 'usKEG', 'usKMI', 'usLMCA',
        'usLTM', 'usLUX', 'usMBRX', 'usMPG', 'usOPXAW', 'usORN', 'usPJT',
        'usPTIE', 'usSAB', 'usSPR', 'usSR', 'usTGEN', 'usTNXP', 'usVBIV',
        'usWMGIZ', 'usXGTIW', 'usMBRX'
    ]

    def __init__(self):
        """被AbuStockBaseWrap替换__init__，即只需读取美股数据到self.df 后续在类装饰器完成"""
        self.df = pd.read_csv(_stock_code_us, index_col=0, dtype=str)

    def __contains__(self, item):
        """成员测试：是否item或item[2:]在self.df.symbol.values中"""
        return item in self.df.symbol.values or (len(item) > 2 and item[2:]
                                                 in self.df.symbol.values)

    def __getitem__(self, key):
        """
            索引获取：两种模式索引获取：
            1. 参数key为df的columns名称，返回self.df[key]，即get df的列
            2. 参数key为股票代码名称，标准化后查询，self.df[self.df.symbol == key]，即get df的行
        """

        if key in self.df.columns:
            # 参数key为df的columns名称，返回self.df[key]，即get df的列
            return self.df[key]

        # get df的行, 即对于股票的详细信息
        if key in self.df.symbol.values:
            return self.df[self.df.symbol == key]
        if len(key) > 2 and key[2:] in self.df.symbol.values:
            return self.df[self.df.symbol == key[2:]]

    def symbol_func(self, df):
        """
        通过df组装支持ABuSymbolPd.make_kl_df使用的symbol，使用('us' + df['symbol']).tolist()
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        # noinspection PyUnresolvedReferences
        return (EMarketTargetType.E_MARKET_TARGET_US.value +
                df['symbol']).tolist()

    def all_symbol(self, index=False):
        """
        获取美股市场中所有股票symbol str对象序列，即美股全市场symbol序列
        :param index: 是否需要返回美股大盘symbol
        :return: 美股全市场symbol序列
        """

        # 过滤AMEX等大盘，etf类型, 只取NASDAQ和NYSE
        df_filter = self.df[(self.df['exchange'] == EMarketSubType.US_OQ.value)
                            |
                            (self.df['exchange'] == EMarketSubType.US_N.value)]
        # 通过symbol_func转换为外部可直接使用ABuSymbolPd.make_kl_df请求的symbol序列
        all_symbol = self.symbol_func(df_filter)
        all_symbol = list(set(all_symbol) - set(AbuSymbolUS.s_unusual_symbol))
        if index:
            # 需要返回大盘symbol
            all_symbol.extend([
                '{}{}'.format(EMarketTargetType.E_MARKET_TARGET_US.value,
                              symbol) for symbol in Symbol.US_INDEX
            ])
        return all_symbol

    def query_symbol_sub_market(self, code, default=EMarketSubType.US_N.value):
        """
        查询股票所在的子市场，即交易所信息, 美股市场默认返回纽交所
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """

        if code in self:
            return self[code].exchange.values[0].upper()
        return default



import sys, os
from IPython.display import clear_output
from IPython.display import display
from ipywidgets import FloatProgress, Text, Box
from ultron.ump.core import env
from ultron.kdutils.decorator import warnings_filter, catch_error
"""多进程下进度条通信socket文件最终名字，这里子进程可以获取g_socket_fn是通过ABuEnvProcess拷贝了主进程全局信息"""
g_socket_fn = None
"""多进程下进度是否显示ui进度，只针对进程间通信类型的进度，有些太频繁的进度显示可以选择关闭"""
g_show_ui_progress = True
"""主进程下用来存贮子进程传递子进程pid为key，进度条对象UIProgress为value"""


def do_clear_output(wait=False):
    """
    模块方法，clear所有的输出，内部针对notebook和命令行输出做区分
    :param wait: 是否同步执行clear操作，透传给IPython.display.clear_output
    """
    if env.g_is_ipython:
        # notebook clear
        clear_output(wait=wait)
    else:
        # cmd clear
        #cmd = 'clear' if env.g_is_mac_os else 'cls'
        cmd = "clear" if True else "cls"
        os.system(cmd)
        # pass


class MulPidProgress(object):
    """多进程进度显示控制类"""

    def __init__(self, total, label, show_progress=True):
        """
        外部使用eg：
        with AbuMulPidProgress(len(self.choice_symbols), 'pick stocks complete') as progress:
            for epoch, target_symbol in enumerate(self.choice_symbols):
                progress.show(epoch + 1)

        :param total: 总任务数量
        :param label: 进度显示label
        """
        self._total = total
        self._label = label
        self.epoch = 0
        self.display_step = 1
        self.progress_widget = None
        self.text_widget = None
        self.progress_box = None
        self.show_progress = show_progress

    # 不管ui进度条有什么问题，也不能影响任务工作的进度执行，反正有文字进度会始终显示
    @catch_error(log=False)
    def init_ui_progress(self):
        """初始化ui进度条"""
        if not self.show_progress:
            return

        if not env.g_is_ipython or self._total < 2:
            return

        if env.g_main_pid == os.getpid():
            # 如果是在主进程下显示那就直接来
            self.progress_widget = FloatProgress(value=0, min=0, max=100)
            self.text_widget = Text('pid={} begin work'.format(os.getpid()))
            self.progress_box = Box([self.text_widget, self.progress_widget])
            display(self.progress_box)
        else:
            if g_show_ui_progress and g_socket_fn is not None:
                # 子进程下通过socket通信将pid给到主进程，主进程创建ui进度条
                ABuOsUtil.socket_send_msg(g_socket_fn,
                                          '{}|init'.format(os.getpid()))

    # 不管ui进度条有什么问题，也不能影响任务工作的进度执行，反正有文字进度会始终显示
    @catch_error(log=False)
    def update_ui_progress(self, ps, ps_text):
        """更新文字进度条"""
        if not self.show_progress:
            return

        if not env.g_is_ipython or self._total < 2:
            return

        if env.g_main_pid == os.getpid():
            # 如果是在主进程下显示那就直接来
            if self.progress_widget is not None:
                self.progress_widget.value = ps
            if self.text_widget is not None:
                self.text_widget.value = ps_text
        else:
            if g_show_ui_progress and g_socket_fn is not None:
                # 子进程下通过socket通信将pid给到主进程，主进程通过pid查找对应的进度条对象后更新进度
                ABuOsUtil.socket_send_msg(
                    g_socket_fn, '{}|{}|{}'.format(os.getpid(), ps, ps_text))

    # 不管ui进度条有什么问题，也不能影响任务工作的进度执行，反正有文字进度会始终显示
    @catch_error(log=False)
    def close_ui_progress(self):
        """关闭ui进度条显示"""
        if not self.show_progress:
            return

        if not env.g_is_ipython or self._total < 2:
            return

        if env.g_main_pid == os.getpid():
            # 如果是在主进程下显示那就直接来
            if self.progress_box is not None:
                self.progress_box.close()
        else:
            if g_show_ui_progress and g_socket_fn is not None:
                # 子进程下通过socket通信将pid给到主进程，主进程通过pid查找对应的进度条对象后关闭对象，且弹出
                ABuOsUtil.socket_send_msg(g_socket_fn,
                                          '{}|close'.format(os.getpid()))

    def __enter__(self):
        """
        以上下文管理器类方式实现__enter__，针对self._total分配self.display_step
        """
        if self.show_progress:
            self.display_step = 1
            if self._total >= 5000:
                self.display_step = 50
            elif self._total >= 3000:
                self.display_step = 30
            elif self._total >= 2000:
                self.display_step = 20
            elif self._total > 1000:
                self.display_step = 10
            elif self._total >= 600:
                self.display_step = 6
            elif self._total >= 300:
                self.display_step = 3
            elif self._total >= 100:
                self.display_step = 2
            elif self._total >= 20:
                self.display_step = 2
            self.epoch = 0
            self.init_ui_progress()
        return self

    def show(self, epoch=None, clear=True):
        """
        进行进度控制显示主方法
        :param epoch: 默认None, 即使用类内部计算的迭代次数进行进度显示
        :param clear: 默认True, 子进程显示新的进度前，先do_clear_output所有输出
        :return:
        """
        if not self.show_progress:
            return

        self.epoch = epoch if epoch is not None else self.epoch + 1
        if self.epoch % self.display_step == 0:
            ps = round(self.epoch / self._total * 100, 2)
            ps = 100 if ps > 100 else ps
            ps_text = "pid:{} {}:{}%".format(os.getpid(), self._label, ps)
            if not env.g_is_ipython or self._total < 2:
                if clear:
                    do_clear_output()
                    # clear_std_output()
                print(ps_text)

            self.update_ui_progress(ps, ps_text)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        以上下文管理器类方式实现__exit__，针对在子进城中的输出显示进度进行do_clear_output扫尾工作
        """
        if not self.show_progress:
            return

        clear = False
        if clear:
            # clear在mac上应该打开, 由于windows某些版本浏览器wait=True会有阻塞情况，如果wait＝False, 有clear之后的风险，
            do_clear_output(wait=True)  # wait 需要同步否则会延迟clear
        else:
            # print("pid:{} done!".format(os.getpid()))
            pass

        self.close_ui_progress()


class Progress(object):
    """单进程（主进程）进度显示控制类"""

    @warnings_filter
    def __init__(self, total, a_progress, label=None):
        """
        外部使用eg：
            progess = Progress(stock_df.shape[0], 0, 'merging {}'.format(m))
            for i, symbol in enumerate(stock_df['symbol']):
                progess.show(i + 1)
        :param total: 总任务数量
        :param a_progress: 初始进度
        :param label: 进度显示label
        """
        self._total = total
        self._progress = a_progress
        self._label = label
        self.f = sys.stdout
        self.progress_widget = None

    def __enter__(self):
        """创建子进程做进度显示"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.f.write('\r')
        if self.progress_widget is not None:
            self.progress_widget.close()

    @property
    def progress(self):
        """property获取self._progress"""
        return self._progress

    @progress.setter
    def progress(self, a_progress):
        """progress.setter设置progress"""
        if a_progress > self._total:
            self._progress = self._total
        elif a_progress < 0:
            self._progress = 0
        else:
            self._progress = a_progress

    def show(self, a_progress=None, ext='', p_format="{}:{}:{}%"):
        """
        进行进度控制显示主方法
        :param ext: 可以添加额外的显示文字，str，默认空字符串
        :param a_progress: 默认None, 即使用类内部计算的迭代次数进行进度显示
        :param p_format: 进度显示格式，默认{}: {}%，即'self._label:round(self._progress / self._total * 100, 2))%'
        """
        self.progress = a_progress if a_progress is not None else self.progress + 1
        ps = round(self._progress / self._total * 100, 2)

        if self._label is not None:
            # 如果初始化label没有就只显示ui进度
            self.f.write('\r')
            self.f.write(p_format.format(self._label, ext, ps))

        if 'IS_IPYTHON' in os.environ and os.environ['IS_IPYTHON']:
            if self.progress_widget is None:
                self.progress_widget = FloatProgress(value=0, min=0, max=100)
                display(self.progress_widget)
            self.progress_widget.value = ps

        # 这样会出现余数结束的情况，还是尽量使用上下文管理器控制结束
        if self._progress == self._total:
            self.f.write('\r')
            if self.progress_widget is not None:
                self.progress_widget.close()
